
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "cell.h"
#include "sif.h"

#ifndef	Float
#define Float	float
#endif

#ifndef	Filter
#define Filter	Chesler
#endif

#ifndef M_PI
#define M_PI	3.14159265358979323846
#endif

extern void	Error(char *msg);
extern double	Filter(int j);

double	Ramachandran(int j)
{
	return (j==0)?1.0/4.0:(j&1)?-1.0/(M_PI*M_PI*(double)j*(double)j):0.0;
}

double	Shepp(int j)
{
	return 2.0/(M_PI*M_PI*(1.0-4.0*(double)j*(double)j));
}

double	Chesler(int j)
{
	if (j==0)
	    return 1.0/8.0-1.0/(2.0*M_PI*M_PI);
	else if (j==1 || j==(-1))
	    return 1.0/16.0-1.0/(2.0*M_PI*M_PI);
	else if (j&1)
	    return -1.0/(2.0*M_PI*M_PI*(double)j*(double)j);
{
	double  j2=(double)j*(double)j,j21=j2-1.0;

	return -(j2+1.0)/(2.0*M_PI*M_PI*j21*j21);
}}

#ifdef	_WIN32
#include <process.h>
#include <windows.h>

#define THREAD_T	HANDLE
#define FUNCTION_T	unsigned
#define FUNCTION_T_CC	FUNCTION_T __stdcall
#define RETURN_VALUE	0

#define INIT_THREAD(T,F,A)	\
	(T=(HANDLE)_beginthreadex(NULL,0,F,(void *)A,0,NULL))==0
#define TERM_THREAD(T)	\
	WaitForSingleObject(T,INFINITE)==WAIT_FAILED || CloseHandle(T)==0
#else
#include <pthread.h>

#define THREAD_T	pthread_t
#define FUNCTION_T	void *
#define FUNCTION_T_CC	FUNCTION_T
#define RETURN_VALUE	NULL

#define INIT_THREAD(T,F,A)	pthread_create(&(T),NULL,F,(void *)A)
#define TERM_THREAD(T)		pthread_join(T,NULL)
#endif

static int	Threads=1;
static THREAD_T	*Thread;

static void	MultiThreading(FUNCTION_T (*body)(),char *name)
{
	size_t	i;

	for (i=1; i<Threads; i++)
	    if (INIT_THREAD(Thread[i],body,i)) {
		(void)fputs(name,stderr);
		Error(" : multi-threading initialization error.");
	    }

	(void)(FUNCTION_T_CC)(*body)((void *)0);

	for (i=1; i<Threads; i++)
	    if (TERM_THREAD(Thread[i])) {
		(void)fputs(name,stderr);
		Error(" : multi-threading termination error.");
	    }
}

#define MULTI_THREADING(FUNCTION)	MultiThreading(FUNCTION,#FUNCTION)

typedef struct {
		double	x,y;
	} Vector;

typedef struct {
		int	j;
		double	r;
	} Struct;

static int	Nr,Nz,N1,N2,N;
static double	**P,**Q,F4CBP;
static Vector	*E,*G,**V;
static Struct	**S4BP;
static Float	***F;

static void	FFT(int sign,Vector *GV)
{
	int	j,k,l,m,p,q;
	Vector	gv,*e,*a,*b;
	double	x,y,z;

	for (j=N2-1, k=N2-2; k>0; k--) {
	    for (l=N1; (j^=l)&l; l>>=1) ;

	    if (j>k) {
		gv=GV[j]; GV[j]=GV[k]; GV[k]=gv;
	    }
	}
	for (j=2, k=1, l=N1, m=N2; m>1; m=l, l>>=1, k=j, j<<=1)
	    for (e=E, p=0; p<k; p++, e+=m) {
		x=e->x; y=e->y*(double)sign;
		for (a=GV+p, q=0; q<l; q++, a+=j) {
		    gv=(*(b=a+k)); b->x=a->x-(z=gv.x*x-gv.y*y); a->x+=z;
				   b->y=a->y-(z=gv.x*y+gv.y*x); a->y+=z;
		}
	    }
}

static FUNCTION_T_CC	Convolution(void *i)
{
	size_t	l;
	int	k,j;
	Vector	*v=V[(size_t)i];
	double	*p,x,y;

	for (l=(size_t)i; l<Nz; l+=Threads) {
	    for (k=0; k<N1; k++) v[k].x=v[k].y=0.0;

	    p=P[l];
	    for (j=0; j<Nr; j++, k++) {
		v[k].x=F4CBP*p[j]; v[k].y=0.0;
	    }
	    for (; k<N2; k++) v[k].x=v[k].y=0.0;

	    FFT(-1,v);

	    for (k=0; k<N2; k++) {
		x=v[k].x;
		y=v[k].y; v[k].x=G[k].x*x-G[k].y*y;
			  v[k].y=G[k].x*y+G[k].y*x;
	    }
	    FFT(1,v);

	    for (j=0; j<Nr; j++) Q[j][l]=v[j].x;
	}
	return RETURN_VALUE;
}

static FUNCTION_T_CC	BackProjection(void *i)
{
	size_t	y;
	int	x,j,z;
	Float	**Fy,*Fyx;
	Struct	*s4bp;
	double	*Q0,*Q1,r0,r1;

	for (y=(size_t)i; y<N; y+=Threads) {
	    Fy=F[y]; s4bp=S4BP[y];
	    for (x=0; x<N; x++)
		if ((j=s4bp[x].j)>=0) {
		    Fyx=Fy[x]; Q0=Q[j]; Q1=Q[j+1]; r1=1.0-(r0=s4bp[x].r);
		    for (z=0; z<Nz; z++) *(Fyx++)+=(*(Q0++)*r1+(*(Q1++))*r0);
		}
	}
	return RETURN_VALUE;
}

#define LEN	2048

int	main(int argc,char **argv)
{
	char	*env,str[LEN],form[LEN],range[LEN],slice[LEN];
	FILE	*file;
	int	Nt,BPS,O,j,i,l,O2,Y,y,O2_Y2,X,x,z,m,lim,norm=0;
	double	Dr,Or,t0,base,step,r,s,c,roof,min,max;
	Cell	*cell;

	if (argc!=7 && argc!=9)
	    Error("usage : xp2tg XP Dr Or t0 {base step} BPS TG/");

	if ((env=getenv("THREADS"))!=NULL &&
	    (Threads=atoi(env))<=0) Error("bad number of THREADS.");

	if (argv[1][0]=='-')
	    file=stdin;
	else
	    if ((file=fopen(argv[1],"rb"))==NULL) Error("file not found.");

	if (fgets(str,LEN,file)==NULL ||
	    sscanf(str,"%d %d %d",&Nr,&Nz,&Nt)!=3 ||
	    Nr<=0 || Nz<=0 || Nt<=0) Error("bad projection data size.");

	if ((Dr=atof(argv[2]))<=0.0) Error("bad interval of detectors.");

	if ((Or=atof(argv[3]))<=0.0 || Or>=(double)(Nr-1))
	    Error("bad center position of detector array.");

	t0=atof(argv[4])*M_PI/180.0;

	BPS=atoi(argv[argc-2]);

	if (argc==9) {
	    base=atof(argv[5]);
	    if ((step=atof(argv[6]))<=0.0) Error("bad CT-value step.");
	}
	else
	    if ((norm=BPS<0)) BPS=(-BPS);

	if (BPS<=0 || BPS>sizeof(Cell)*8) Error("bad BPS of CT image.");

	for (N1=1; N1<Nr; N1<<=1) ; N2=N1<<1;

	N=(O=(int)floor(((r=(double)(Nr-1)-Or)>Or)?Or:r))*2+1;

	if ((Thread=(THREAD_T *)malloc(sizeof(THREAD_T)*Threads))==NULL ||
	    (E=(Vector *)malloc(sizeof(Vector)*N2))==NULL ||
	    (G=(Vector *)malloc(sizeof(Vector)*N2))==NULL ||
	    (V=(Vector **)malloc(sizeof(Vector *)*Threads))==NULL ||
	    (*V=(Vector *)malloc(sizeof(Vector)*Threads*N2))==NULL ||
	    (P=(double **)malloc(sizeof(double *)*Nz))==NULL ||
	    (*P=(double *)malloc(sizeof(double)*Nz*Nr))==NULL ||
	    (Q=(double **)malloc(sizeof(double *)*(Nr+1)))==NULL ||
	    (*Q=(double *)malloc(sizeof(double)*Nr*Nz))==NULL ||
	    (S4BP=(Struct **)malloc(sizeof(Struct *)*N))==NULL ||
	    (*S4BP=(Struct *)malloc(sizeof(Struct)*N*N))==NULL ||
	    (F=(Float ***)malloc(sizeof(Float **)*N))==NULL ||
	    (*F=(Float **)malloc(sizeof(Float *)*N*N))==NULL ||
	    (**F=(Float *)malloc(sizeof(Float)*N*N*Nz))==NULL)
	    Error("memory allocation error.");

	F4CBP=1.0/Dr/(double)N2*M_PI/(double)Nt;

	for (j=0; j<N2; j++) {
	    sincos(M_PI*(double)j/(double)N2,&(E[j].y),&(E[j].x));

	    G[j].x=Filter(j-N1); G[j].y=0.0;
	}
	FFT(-1,G);

	for (i=1; i<Threads; i++) V[i]=V[i-1]+N2;

	for (l=1; l<Nz; l++) P[l]=P[l-1]+Nr;

	Q[Nr]=Q[0]; for (j=1; j<Nr; j++) Q[j]=Q[j-1]+Nz;

	O2=O*O;
	for (Y=(-O), y=0; y<N; y++, Y++) {
	    if (y>0) {
		S4BP[y]=S4BP[y-1]+N; F[y]=F[y-1]+N;
	    }
	    O2_Y2=O2-Y*Y;
	    for (X=(-O), x=0; x<N; x++, X++) {
		S4BP[y][x].j=(O2_Y2>=X*X)-1;

		if (y>0 || x>0) F[y][x]=F[y][x-1]+Nz;

		for (z=0; z<Nz; z++) F[y][x][z]=0.0;
	    }
	}
	for (m=0; m<Nt; m++) {
	    if (fread(*P,sizeof(double),(size_t)Nz*Nr,file)!=(size_t)Nr*Nz)
		Error("projection data read error.");

	    MULTI_THREADING(Convolution);

	    sincos(t0+M_PI*(double)m/(double)Nt,&s,&c);
	    for (Y=(-O), y=0; y<N; y++, Y++)
	    for (X=(-O), x=0; x<N; x++, X++)
		if (S4BP[y][x].j>=0) {
		    S4BP[y][x].j=j=(int)(r=Or+c*(double)X+s*(double)Y);
		    S4BP[y][x].r=r-(double)j;
		}

	    MULTI_THREADING(BackProjection);
	}
	if (file!=stdin) (void)fclose(file);

	i=1; for (z=Nz-1; z>=10; z/=10) ++i; (void)sprintf(form,"%%0%dd",i);

	lim=(1<<BPS)-1;

	if (argc==9)
	    (void)sprintf(range,"%lf\t%lf",base,roof=base+step*(double)lim);
	else if (norm) {
	    min=max=F[0][0][0];
	    for (y=0; y<N ; y++)
	    for (x=0; x<N ; x++)
	    for (z=0; z<Nz; z++)
		if (F[y][x][z]<min) min=F[y][x][z]; else
		if (F[y][x][z]>max) max=F[y][x][z];

	    (void)sprintf(range,"%lf\t%lf",min,max);
	    base=min; step=(min<max)?(max-min)/(double)lim:1.0;

	    (void)fputs(range,stderr); (void)fputc('\n',stderr);
	}
	for (z=0; z<Nz; z++) {
	    min=max=F[0][0][z];
	    for (y=0; y<N; y++)
	    for (x=0; x<N; x++)
		if (F[y][x][z]<min) min=F[y][x][z]; else
		if (F[y][x][z]>max) max=F[y][x][z];

	    (void)sprintf(slice,form,z);
	    (void)printf("%s\t%lf\t%lf\n",slice,min,max);

	    if (argc==9)
		for (y=0; y<N; y++) {
		    cell=(Cell *)S4BP[N-1-y];
		    for (x=0; x<N; x++)
			cell[x]=(F[y][x][z]<base)?0:
				(F[y][x][z]>roof)?(Cell)lim:
				(Cell)((F[y][x][z]-base)/step+0.5);
		}
	    else {
		if (!norm) {
		    (void)sprintf(range,"%lf\t%lf",min,max);
		    base=min; step=(min<max)?(max-min)/(double)lim:1.0;
		}
		for (y=0; y<N; y++) {
		    cell=(Cell *)S4BP[N-1-y];
		    for (x=0; x<N; x++)
			cell[x]=(Cell)((F[y][x][z]-base)/step+0.5);
		}
	    }
	    (void)sprintf(str,"%s/%s.tif",argv[argc-1],slice);
	    StoreImageFile(str,N,N,BPS,(Cell **)S4BP,range);
	}
	free(**F); free(*F); free(F);
	free(*S4BP); free(S4BP); free(*Q); free(Q);
	free(*P); free(P); free(*V); free(V); free(G); free(E); free(Thread);

	return 0;
}
